package kms

import (
	"fmt"
	"sync"

	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
)

type Status int

const (
	RESERVED = iota
	AVAILABLE
	UNKNOWN
	USED
)

// holds a single ready to bit key, length can be configured
type kmsKSElement struct {
	keyID  uuid.UUID
	key    []byte // a key
	status Status
}

type kmsKeyStore struct {
	keyStoreMutex sync.Mutex
	keyStore      map[uuid.UUID]*kmsKSElement
	keySingleSize uint // the size of a single key, given as unit of bits
}

func NewKmsKeyStore(desiredkeySingleSizeLength uint) *kmsKeyStore {
	return &kmsKeyStore{
		// NOTE: could be useful to use something like a ordered map here
		keyStore:      make(map[uuid.UUID]*kmsKSElement),
		keySingleSize: desiredkeySingleSizeLength,
	}
}

func (ks *kmsKeyStore) addKey(keyId uuid.UUID, keyToadd []byte) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	// test for collisions
	if _, notThere := ks.keyStore[keyId]; notThere {
		log.Errorf("Whop: addKey collission of key id %s", keyId)
		return
	}

	newKeyElement := &kmsKSElement{
		keyID:  keyId,
		key:    keyToadd,
		status: AVAILABLE,
	}
	// ok to add
	ks.keyStore[newKeyElement.keyID] = newKeyElement
}

func (ks *kmsKeyStore) GetKey() (*kmsKSElement, error) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	// TODO: if we want random here, then we have to take a different approach
	for _, key := range ks.keyStore {
		if key.status == AVAILABLE {
			// change status of key to reserved
			key.status = RESERVED
			return key, nil
		}
	}
	return nil, fmt.Errorf("could not find a key to use")
}

func (ks *kmsKeyStore) GetKeyWithID(id uuid.UUID) (*kmsKSElement, error) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	key, ok := ks.keyStore[id]
	if !ok {
		return nil, fmt.Errorf("Could not find a key for id: %v", id)
	}
	if key.status == AVAILABLE {
		// change status of key to reserved
		key.status = RESERVED
		return key, nil
	}
	return nil, fmt.Errorf("Key with index: %v is already reserved.", id)
}

func (ks *kmsKeyStore) DeleteKey(keyId uuid.UUID) {
	ks.keyStoreMutex.Lock()
	delete(ks.keyStore, keyId)
	ks.keyStoreMutex.Unlock()
}
