package kms

import (
	"context"
	"encoding/base64"
	"io/ioutil"
	"time"

	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"

	etsi14 "code.fbi.h-da.de/danet/ekms/api/go/rest/etsi/client"
	pb "code.fbi.h-da.de/danet/proto-kms/api/gen/proto/go/kmsintercom"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"
)

type kmsTalkerServer struct {
	pb.UnimplementedKmsTalkerServer
	keyNegotiationMap map[uuid.UUID]*kmsKSElement
	eKMS              *EKMS
}

// This must somehow find out and agree to a specific key lenght
func (s *kmsTalkerServer) InterComCapabilities(ctx context.Context, in *pb.InterComCapabilitiesRequest) (capReply *pb.InterComCapabilitiesReply, err error) {
	log.Debugf("Received: %v", in.GetMyKmsName())

	// TODO: Call to ksp := NewKmsKeyStore(<desired-size-of-each-key-in-bits)
	// this to be stored in the serving QLE QuantumElement struct under keyStorePeer
	// Further, the KMS peers have to agree on a ready-to-be-used keyBulk based on the bulkId
	// This requires to go through the rawBulkKeys of type QuantumElement and lookup a bulkId both side do know Once agreed upon on keyBulk, this here has to call KeyChopper of the actual NewKmsKeyStore

	return &pb.InterComCapabilitiesReply{
		PeerKmsName: "whatever",
	}, nil
}

func (s *kmsTalkerServer) KeyIdNotification(ctx context.Context, in *pb.KeyIdNotificationRequest) (*pb.KeyIdNotificationResponse, error) {
	// check if a peer exists
	peer, ok := s.eKMS.KmsPeers[in.GetInterComAddr()]
	if !ok {
		//TODO: proper error message
		return nil, status.Error(codes.Internal, "peer does not exist")
	}

	eqm, ok := peer.servingQuantumModul.(*ETSI014HTTPQuantumModule)
	if !ok {
		return nil, status.Error(codes.Internal, "expected etsi014 quantum module")
	}

	etsi14KeyIds := make([]etsi14.KeyIDsRequestKeyIDsInner, len(in.KeyIds))
	for i, keyid := range in.KeyIds {
		etsi14KeyIds[i] = etsi14.KeyIDsRequestKeyIDsInner{
			KeyID: keyid,
		}
	}

	keyContainer, resp, err := eqm.client.GetKeyWithIdPost(eqm.masterSAEID, etsi14KeyIds)
	if err != nil {
		return nil, err
	}

	defer resp.Body.Close()

	// TODO: add proper status code handling
	if resp.StatusCode != 200 {
		body, err := ioutil.ReadAll(resp.Body)
		if err != nil {
			return nil, err
		}
		return nil, status.Error(codes.Aborted, string(body))
	}

	// TODO: could be run in go routine
	if err := addETSIKeysToKeystore(eqm.keyStore, keyContainer.GetKeys()); err != nil {
		return nil, status.Error(codes.Internal, "expected etsi014 quantum module")
	}

	return &pb.KeyIdNotificationResponse{
		Timestamp: time.Now().Unix(),
	}, nil
}

// TODO: should be removed as soon as the emulated quantum module has been
// changed; is specific for emulated quantum module
func (s *kmsTalkerServer) SyncQkdBulk(ctx context.Context, in *pb.SyncQkdBulkRequest) (*pb.SyncQkdBulkResponse, error) {
	// NOTE: with "google.golang.org/grpc/peer" it would be possible to get the client ip directly

	// check if a peer exists
	peer, ok := s.eKMS.KmsPeers[in.GetInterComAddr()]
	if !ok {
		//TODO: proper error message
		return nil, status.Errorf(codes.Internal, "peer does not exist")
	}

	eqm, ok := peer.servingQuantumModul.(*EmulatedQuantumModule)
	if !ok {
		return nil, status.Errorf(codes.Internal, "expected emulated quantum module")
	}

	for _, bulkId := range in.BulkId {
		if _, ok := eqm.rawBulkKeys[bulkId]; ok {
			return &pb.SyncQkdBulkResponse{
				Timestamp: time.Now().Unix(),
				BulkId:    bulkId,
			}, nil
		}
	}

	return nil, status.Errorf(codes.Unimplemented, "method SyncQkdBulk not implemented")
}

func (s *kmsTalkerServer) SyncKeyIdsForBulk(ctx context.Context, in *pb.SyncKeyIdsForBulkRequest) (*pb.SyncKeyIdsForBulkResponse, error) {
	peer, ok := s.eKMS.KmsPeers[in.GetInterComAddr()]
	if !ok {
		return nil, status.Errorf(codes.Internal, "For address: %s, no peer exists", in.GetInterComAddr())
	}

	eqm, ok := peer.servingQuantumModul.(*EmulatedQuantumModule)
	if !ok {
		return nil, status.Errorf(codes.Internal, "expected a emulated quantum module")
	}

	eqm.rawBulkKeysMutex.Lock()
	bulk, ok := eqm.rawBulkKeys[in.GetBulkId()]
	if !ok {
		return nil, status.Errorf(codes.Internal, "bulk key with id: %d not found", in.GetBulkId())
	}
	eqm.rawBulkKeysMutex.Unlock()

	eqm.keyStore = NewKmsKeyStore(256)
	_, keyData, err := eqm.KeyChopper(bulk, in.GetKeyId())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%v", err)
	}

	for keyId, key := range keyData {
		eqm.keyStore.addKey(keyId, key)
	}

	eqm.rawBulkKeysMutex.Lock()
	delete(eqm.rawBulkKeys, in.GetBulkId())
	eqm.rawBulkKeysMutex.Unlock()

	// update the peer status to up
	peer.peerStatus = KmsPeerUp
	// Send notification about change
	if peer.externalNotifierKMSPeer != nil {
		peer.externalNotifierKMSPeer <- peer.tcpSocketStr
	}

	return &pb.SyncKeyIdsForBulkResponse{
		Timestamp: time.Now().Unix(),
	}, nil
}

func (s *kmsTalkerServer) InterComTransportKeyNegotiation(ctx context.Context, in *pb.InterComTransportKeyNegotiationRequest) (capReply *pb.InterComTransportKeyNegotiationResponse, err error) {
	// NOTE: For the current prototype it is assumed that a negotiation request
	// is always valid. In the future an incoming negotiation request should
	// also check for a suitable forwarding assignment from the controller.
	keyID, err := uuid.Parse(in.GetKeyToUse())
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "key with id: %s can not be parsed to uuid", in.GetKeyToUse())
	}
	pathId, err := uuid.Parse(in.GetPathID())
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "path id: %s can not be parsed to uuid", in.GetPathID())
	}

	route, ok := s.eKMS.routingTable[pathId]
	if !ok {
		return nil, status.Errorf(codes.Internal, "There is no route for the given pathID: %s .", in.PathID)
	}

	//TODO: This limits a pathId so that it is only possible to send one single
	//payload.
	//if _, ok := s.keyNegotiationMap[pathId]; ok {
	//	return nil, status.Errorf(codes.Internal, "A transport key for pathID: %s has already been negotiated.", in.PathID)
	//}

	quantumModuleKeyStore := route.Previous.servingQuantumModul.KeyStore()

	key, err := quantumModuleKeyStore.GetKeyWithID(keyID)
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%v", err)
	}

	// TODO: mutex
	s.keyNegotiationMap[pathId] = key

	quantumModuleKeyStore.DeleteKey(keyID)

	return &pb.InterComTransportKeyNegotiationResponse{Timestamp: time.Now().Unix()}, nil
}

func (s *kmsTalkerServer) KeyForwarding(ctx context.Context, in *pb.KeyForwardingRequest) (capReply *pb.KeyForwardingResponse, err error) {
	pathId, err := uuid.Parse(in.PathId)
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	decryptKey, ok := s.keyNegotiationMap[pathId]
	if !ok {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	route, ok := s.eKMS.routingTable[pathId]
	if !ok {
		return nil, status.Errorf(codes.Internal, "There is no route for the given pathID: %s .", in.PathId)
	}

	payloadAsByte, err := base64.StdEncoding.DecodeString(in.Payload)
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%s", err)
	}

	decryptedPayload, err := route.Previous.et.Decrypt(payloadAsByte, decryptKey.key)
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%s", err)
	}

	log.Infof("%s received a payload: %s, from %s", s.eKMS.kmsName, decryptedPayload, route.Previous.tcpSocketStr)

	if route.Next != nil {
		log.Infof("%s forwards payload to : %s", s.eKMS.kmsName, route.Next.tcpSocketStr)
		go route.Next.SendPayload(decryptedPayload, pathId)
	} else {
		log.Infof("%s received the final payload: %s", s.eKMS.kmsName, string(decryptedPayload))
	}

	//// NOTE: For demo purpose only
	//json, err := json.Marshal(KMSInfo{
	//	Name:             s.eKMS.kmsName,
	//	EncryptedMessage: in.Payload,
	//	DecryptedMessage: string(decryptedPayload),
	//	Key:              base64.StdEncoding.EncodeToString(decryptKey.key),
	//})
	//if err != nil {
	//	log.Println("Failed to marshal: ", err)
	//}

	//err = sendKmsInfoMessage("http://172.20.10.21:4000/kmsinfo", json)
	//if err != nil {
	//	log.Println("Failed to send KMS info message: ", err)
	//}

	return &pb.KeyForwardingResponse{Timestamp: time.Now().Unix()}, nil
}
