package kms

import (
	"context"
	"encoding/base64"
	"fmt"
	"net"
	"time"

	pbIC "code.fbi.h-da.de/danet/proto-kms/api/gen/proto/go/kmsintercom"
	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
)

type KmsPeerStatus int16

const (
	KmsPeerUp      KmsPeerStatus = iota // peer was seen up lately
	KmsPeerDown                         // peer not reachable anymore
	KmsPeerPending                      // peer was added and has to be succesfully contacted once
	KmsPeerUnknown                      // not known, not initialized
)

type kmsPeerInfo interface {
	GetKmsPeerStatus() KmsPeerStatus
	GetKmsPeerId() uuid.UUID
	GetKmsPeerQkdiId() uint32
	KmsPeerKeyInit()
}

type kmsPeer struct {
	externalNotifierKMSPeer chan string
	peerClient              pbIC.KmsTalkerClient
	peerStatus              KmsPeerStatus
	peerKmsId               string // NOTE: might be changed in the future
	interComAddr            string
	servingQuantumModul     QuantumModule
	tcpSocket               net.TCPAddr // the IP address and TCP port (aka socket) of the kms peer
	tcpSocketStr            string      // string rep. of tcpSocket
	et                      CryptoAlgorithm
	name                    string    // the name of the kms peer
	id                      uuid.UUID // uuid of the peer
	quit                    chan bool // cancel the peer goroutine
}

func NewKmsPeer(peerKmsId string, servQM QuantumModule, tcpSocketStr string, interComAddr string, in chan string) (*kmsPeer, error) {
	if servQM.Peer() != nil {
		return nil, fmt.Errorf("QuantumModule with ID: , already has a peer", servQM.ID())
	}
	newPeerConn, err := grpc.Dial(tcpSocketStr, grpc.WithTransportCredentials(insecure.NewCredentials()))
	if err != nil {
		return nil, err
	}
	peerClient := pbIC.NewKmsTalkerClient(newPeerConn)

	peer := &kmsPeer{
		externalNotifierKMSPeer: in,
		// TODO: rename to client
		peerClient: peerClient,
		// TODO: change this, only for demo purposes
		peerStatus: KmsPeerUp,
		peerKmsId:  peerKmsId,
		// TODO: move this into a config
		interComAddr:        interComAddr,
		servingQuantumModul: servQM,
		tcpSocketStr:        tcpSocketStr,
		et:                  NewAES(),
		id:                  uuid.New(),
		quit:                make(chan bool),
	}

	servQM.SetPeer(peer)

	if err := servQM.Initialize(); err != nil {
		return nil, err
	}

	return peer, nil
}

func (ph *kmsPeer) SyncBulkKeys() error {
	return ph.servingQuantumModul.Sync()
}

// TransportKeyNegotiation sends a request for a transport key negotiation to
// the peer KMS. The key that was agreed upon, is used to create a encrypted
// payload.
func (ph *kmsPeer) TransportKeyNegotiation() error {
	//ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	//req, err := ph.peerClient.InterComTransportKeyNegotiation()
	return nil
}

func (ph *kmsPeer) SendPayload(payload []byte, pathId uuid.UUID) error {
	// NOTE: It should be assumed that there are keys available if we try to
	// send.
	if len(ph.servingQuantumModul.KeyStore().keyStore) == 0 {
		log.Debug("Syncing bulk keys before sending the payload")
		err := ph.SyncBulkKeys()
		if err != nil {
			return err
		}
	}

	// TODO: Return a message if keys are empty

	// select a key from key store
	key, err := ph.servingQuantumModul.KeyStore().GetKey()
	if err != nil {
		return err
	}

	// Start the negotiation process for a transport key
	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()
	_, err = ph.peerClient.InterComTransportKeyNegotiation(ctx, &pbIC.InterComTransportKeyNegotiationRequest{
		Timestamp: time.Now().Unix(),
		PathID:    pathId.String(),
		KeyToUse:  key.keyID.String(),
	})
	if err != nil {
		return err
	}

	ph.servingQuantumModul.KeyStore().DeleteKey(key.keyID)

	// TODO: would be better to update the index counter here (to keep it
	// synchronized).

	encryptedPayload, err := ph.et.Encrypt(payload, key.key)
	if err != nil {
		return err
	}

	encryptedPayloadAsString := base64.StdEncoding.EncodeToString(encryptedPayload)

	log.Infof("Sent encrypted Payload: %s", encryptedPayloadAsString)

	ctx2, cancel2 := context.WithTimeout(context.Background(), time.Second)
	defer cancel2()
	_, err = ph.peerClient.KeyForwarding(ctx2, &pbIC.KeyForwardingRequest{
		Timestamp: time.Now().Unix(),
		PathId:    pathId.String(),
		Payload:   encryptedPayloadAsString,
	})
	if err != nil {
		return err
	}

	return nil
}

func (ph *kmsPeer) GetKmsPeerStatus() KmsPeerStatus {
	return ph.peerStatus
}

func (ph *kmsPeer) GetKmsPeerId() uuid.UUID {
	return ph.id
}
